import os
import numpy as np
import math
import pandas as pd
import cv2
import pickle
import colorsys
import pyqtgraph as pg
from scipy.interpolate import interp1d, splprep, splev
from .uuuuuu import rotation_z, rotation_x, rotation_y


def get_closest_point_to_line(p0, r, p):
    ap = p - p0
    res = p0 + np.dot(ap, r) * r
    return res


def line_fit_2d(points_2d, image=None):
    msg = None
    points = np.asarray(points_2d)
    sort_order = np.argsort(points[:, 1])
    points = points[sort_order, :]
    avg = np.mean(points, 0)
    subtracted = points - avg
    u, s, vh = np.linalg.svd(subtracted)
    direction = vh[0, :] / np.linalg.norm(vh[0, :])
    p1 = points[0, :]
    p2 = points[-1, :]
    sp = get_closest_point_to_line(avg, direction, p1)
    ep = get_closest_point_to_line(avg, direction, p2)

    if image is not None:
        direction = sp - ep
        direction = direction / np.linalg.norm(direction)

        valid_ind = np.where(image[:, int(sp[0])] != 0)[0]
        if len(valid_ind) > 0:
            valid_ind = valid_ind[0]
        else:
            msg = "Something went wrong for the probe location. Please contact maintainers."

        if int(sp[1]) < valid_ind:
            sign_flag = -1
        elif int(sp[1]) > valid_ind:
            sign_flag = 1
        else:
            sign_flag = 0

        stop_steps = 0
        temp = sp + 0
        while int(temp[1]) != valid_ind:
            stop_steps += 1
            temp = sp + sign_flag * stop_steps * direction
            valid_ind = np.where(image[:, int(temp[0])] != 0)[0]
            if len(valid_ind) > 0:
                valid_ind = valid_ind[0]
            else:
                msg = "Something went wrong for the probe location. Please contact maintainers."
                break

        sp = sp + sign_flag * stop_steps * direction

    p2 = np.array([sp, ep])
    return p2, msg


def line_fit(points):
    points = np.asarray(points)
    sort_order = np.argsort(points[:, 2])[::-1]
    points = points[sort_order, :]
    avg = np.mean(points, 0)
    substracted = points - avg
    u, s, vh = np.linalg.svd(substracted)
    direction = vh[0, :] / np.linalg.norm(vh[0, :])
    p1 = points[0, :]
    p2 = points[-1, :]
    sp = get_closest_point_to_line(avg, direction, p1)
    ep = get_closest_point_to_line(avg, direction, p2)
    return sp, ep, avg, direction


def get_angles(direction):
    direction = direction / np.linalg.norm(direction)

    vertical_vec = np.array([0, 0, 1])
    ap_proj = direction.copy()
    ap_proj[0] = 0
    ap_proj = ap_proj / np.linalg.norm(ap_proj)
    ml_proj = direction.copy()
    ml_proj[1] = 0
    ml_proj = ml_proj / np.linalg.norm(ml_proj)

    ap_val = math.acos(np.max([np.min([np.dot(ap_proj, vertical_vec), 1]), -1]))
    ml_val = math.acos(np.max([np.min([np.dot(ml_proj, vertical_vec), 1]), -1]))

    ap_angle = np.degrees(ap_val)
    ml_angle = np.degrees(ml_val)

    if ap_angle > 90:
        ap_angle = 180 - ap_angle
    if ml_angle > 90:
        ml_angle = 180 - ml_angle

    return ap_angle, ml_angle


def get_tilt_info(sp, ep):
    if ep[1] < sp[1]:
        ap_tilt = "posterior"
    elif ep[1] > sp[1]:
        ap_tilt = "anterior"
    else:
        ap_tilt = "no tilt"

    if sp[0] > 0:
        if ep[0] < sp[0]:
            ml_tilt = "medial"
        elif ep[0] > sp[0]:
            ml_tilt = "lateral"
        else:
            ml_tilt = "no tilt"
    elif sp[0] < 0:
        if ep[0] < sp[0]:
            ml_tilt = "lateral"
        elif ep[0] > sp[0]:
            ml_tilt = "medial"
        else:
            ml_tilt = "no tilt"
    else:
        if ep[0] != sp[0]:
            ml_tilt = "lateral"
        else:
            ml_tilt = "no tilt"
    return ap_tilt, ml_tilt


def get_tilt_sign(sp, ep):
    if ep[1] < sp[1]:
        ap_tilt = -1
    elif ep[1] > sp[1]:
        ap_tilt = 1
    else:
        ap_tilt = 0

    if ep[0] < sp[0]:
        ml_tilt = -1
    elif ep[0] > sp[0]:
        ml_tilt = 1
    else:
        ml_tilt = 0
    return ap_tilt, ml_tilt


def pandas_to_str(label_name, label_ano, length, channels):
    df = pd.DataFrame(
        {
            "Brain Regions": label_name,
            "Ano": label_ano,
            "Probe Length": length,
            "Probe Channels": channels,
        }
    )
    return df.to_string(col_space=30, justify="justify")


def correct_start_pnt(
    label_data, start_pnt, start_vox, direction, bregma, verbose=False
):
    error_index = 0
    # print('start_vox', start_vox)
    direction = direction / np.linalg.norm(direction)
    check_vec = label_data[int(start_vox[0]), int(start_vox[1]), :]
    top_vox = np.where(check_vec != 0)[0]
    # print(np.where(check_vec != 0))
    if len(top_vox) == 0:
        print("wrong")
    else:
        top_vox = top_vox[-1]
    check_vox = start_vox.astype(int)
    new_sp = start_pnt.copy()

    # if int(start_vox[2]) < top_vox:
    #     steps = 0
    #     while label_data[check_vox[0], check_vox[1], check_vox[2]] != 0:
    #         steps += 1
    #         new_sp = start_pnt - steps * direction
    #         check_vox = new_sp + bregma
    #         check_vox = check_vox.astype(int)
    #         if steps == 1000:
    #             raise Exception('higher limit, please contact maintainer')
    # elif int(start_vox[2]) > top_vox:
    #     steps = 0
    #     while label_data[check_vox[0], check_vox[1], check_vox[2]] == 0:
    #         steps += 1
    #         new_sp = start_pnt + steps * direction
    #         check_vox = new_sp + bregma
    #         check_vox = check_vox.astype(int)
    #         if steps == 1000:
    #             raise Exception('higher limit, please contact maintainer')
    # else:
    #     new_sp = start_pnt
    enter_label = label_data[check_vox[0], check_vox[1], check_vox[2]]
    z_diff = new_sp[2] + bregma[2] - top_vox
    if z_diff >= 1:
        sign_flag = -1
        if enter_label != 0:
            error_index = 10
            return new_sp, error_index
    elif z_diff < 0:
        sign_flag = 1
        if enter_label == 0:
            error_index = 11
            return new_sp, error_index
    else:
        sign_flag = 0

    # ToDo: the following code only works when there is no 0 label inside the brain (label data)
    # that means label 0 only indicates the area outside the brain
    # if some atlas uses label 0 to indicate the unspecified area inside the brain, this code may have problems
    # should make a pseudo label index (xxxx) inside the brain for unspecified regions ===> need to re-process atlas

    stop_steps = 0
    if sign_flag != 0:
        enter_condition = (enter_label == 0) if z_diff >= 1 else (enter_label != 0)
        while enter_condition:
            stop_steps += 1
            new_sp = start_pnt - sign_flag * stop_steps * direction
            check_vox = new_sp + bregma
            check_vox = check_vox.astype(int)
            if np.any(check_vox >= np.ravel(label_data.shape)) or np.any(check_vox) < 0:
                error_index = 12
                break
            enter_label = label_data[check_vox[0], check_vox[1], check_vox[2]]
            enter_condition = (enter_label == 0) if z_diff >= 1 else (enter_label != 0)
        if error_index != 0:
            return new_sp, error_index

        new_sp = (
            start_pnt - sign_flag * (stop_steps - 1) * direction
            if z_diff < 0
            else new_sp
        )
        if verbose:
            print("correct enter pnt with {} steps".format(stop_steps))
            # print('old sp', start_pnt)
            # print('new pc_sp', new_sp)

    return new_sp, error_index


def correct_end_point(
    sp, ep, direction, vox_size, tip_length, max_probe_length, verbose=False
):
    probe_length_with_tip = np.sqrt(np.sum((sp - ep) ** 2)) * vox_size  # in um
    probe_length_without_tip = probe_length_with_tip - tip_length  # in um
    if max_probe_length is not None:
        if probe_length_with_tip > max_probe_length:
            probe_length_with_tip = max_probe_length
            probe_length_without_tip = probe_length_with_tip - tip_length
            new_ep = sp + direction * probe_length_with_tip / vox_size
        else:
            new_ep = ep
    else:
        new_ep = ep
    if verbose:
        print("old pc_ep", ep)
        print("corrected pc_ep", new_ep)
        print("old probe length with tip", probe_length_with_tip)
        print("corrected probe length with tip", probe_length_with_tip)
    return new_ep, probe_length_with_tip, probe_length_without_tip


def check_parallel_to_z(direction):
    is_parallel = False
    if abs(abs(direction[2]) - 1) < 1e-6:
        is_parallel = True
    return is_parallel


def angle_between_2vectors(vector1, vector2):
    unit_vector_1 = vector1 / np.linalg.norm(vector1)
    unit_vector_2 = vector2 / np.linalg.norm(vector2)
    dot_product = np.dot(unit_vector_1, unit_vector_2)
    angle = np.arccos(dot_product)
    return angle


def get_probe_info(probe_type):
    if probe_type == 0:
        tip_length = 175
        channel_size = 20
        channel_number_in_banks = (384, 384, 192)
    elif probe_type == 1:
        tip_length = 175
        channel_size = 15
        channel_number_in_banks = (384, 384, 192)
    else:
        tip_length = 0
        channel_size = None
        channel_number_in_banks = None
    return tip_length, channel_size, channel_number_in_banks


def group_labels(fine_label_mat, verbose=False):
    n_row, n_column = fine_label_mat.shape
    group_mat = np.zeros(fine_label_mat.shape)
    group_mat[:] = np.nan
    group_id = 0
    group_id_label = []

    previous_row = np.zeros(n_column)
    previous_row[:] = np.nan
    for base_level in range(0, n_row):
        check_label = fine_label_mat[base_level]
        unique_label = np.unique(check_label)
        for da_label in unique_label:
            if da_label in previous_row:
                continue
            add_row = 0
            row_index = add_row + base_level
            label_check = fine_label_mat[row_index] == da_label
            group_mat[row_index, label_check] = group_id
            while np.any(label_check):
                add_row += 1
                row_index = add_row + base_level
                if row_index == n_row:
                    break
                label_check = fine_label_mat[row_index] == da_label
                group_mat[row_index, label_check] = group_id
            group_id_label.append(da_label)
            group_id += 1
        previous_row = fine_label_mat[base_level]

    if verbose:
        print("group mat nan index", np.where(np.isnan(group_mat)))
        print("group id label", group_id_label)
        print("group_mat")
        indexs = np.arange(0, len(group_mat), 20)
        for i in range(len(indexs) - 1):
            print(group_mat[indexs[i] : indexs[i + 1]])
        print(group_mat[indexs[-1] : len(group_mat)])
        print(len(group_mat))

    return group_mat, group_id_label


def get_column_grouped_info(group_column, column_bound):
    diff_labels = np.where(np.diff(group_column) != 0)[0]
    if len(diff_labels) != 0:
        change_index = np.append(np.array([0]), diff_labels + 1)
        group_id = group_column[change_index]
        start_loc = column_bound[change_index]
        end_loc = column_bound[
            np.append(change_index[1:], np.array([len(group_column)]))
        ]
    else:
        group_id = group_column[0]
        start_loc = column_bound[0]
        end_loc = column_bound[-1]

    group_column_length = np.abs(end_loc - start_loc)

    return group_id, start_loc, end_loc, group_column_length


def get_vis_data(group_mat, column_loc, sites_loc, sites_line_count, vox_size):
    n_column = len(column_loc)

    p_bounds = np.array([0])
    p_bounds = np.append(
        p_bounds, column_loc[0][:-1, 0] + np.diff(column_loc[0][:, 0]) * 0.5
    )
    p_bounds = np.append(p_bounds, column_loc[0][-1, 0])

    vis_data = []
    column_group_length = []
    column_n_sites = []
    for i in range(n_column):
        sites_column = sites_loc[i][:, 0]
        # print(sites_column)
        if sites_line_count is None:
            line_counts = np.ones(len(sites_column))
        else:
            line_counts = np.ravel(sites_line_count)
        group_column = group_mat[:, i]
        group_id, start_loc, end_loc, gc_length = get_column_grouped_info(
            group_column, p_bounds
        )
        n_sites_group = []
        for j in range(len(group_id)):
            valid_sites_lines = np.logical_and(
                sites_column < end_loc[j], sites_column >= start_loc[j]
            )
            n_sites_group.append(np.sum(valid_sites_lines * line_counts))
        # print('n_sites_group', n_sites_group)
        vis_column = {
            "group_id": group_id,
            "start_loc": start_loc / vox_size,
            "end_loc": end_loc / vox_size,
            "sites": sites_column / vox_size,
        }
        column_group_length.append(gc_length)
        column_n_sites.append(np.ravel(n_sites_group))
        vis_data.append(vis_column)
    # print(vis_data)
    n_column = len(vis_data)
    unique_groups = np.unique(group_mat)
    group_length = []
    group_n_sites = []
    for group_ind in unique_groups:
        temp = []
        temp_sites = []
        for i in range(n_column):
            valid_ind = np.where(vis_data[i]["group_id"] == group_ind)[0]
            if len(valid_ind) != 0:
                valid_length = np.sum(column_group_length[i][valid_ind])
                valid_sites = np.sum(column_n_sites[i][valid_ind])
                temp.append(valid_length)
                temp_sites.append(valid_sites)
        group_length.append(np.sum(temp) / n_column)
        group_n_sites.append(np.sum(temp_sites))

    text_loc = []
    for group_ind in unique_groups:
        inds = np.where(group_mat == group_ind)[0]
        low_level = np.min(inds)
        high_level = np.max(inds)
        text_loc.append(
            (
                p_bounds[low_level]
                + 0.7 * (p_bounds[high_level + 1] - p_bounds[low_level])
            )
            / vox_size
        )
    return vis_data, group_length, group_n_sites, text_loc


def get_n_sites_in_region(gr_start, gr_end, plot_sites_column):
    n_groups, n_column = gr_start.shape

    region_site_num = []
    for i in range(n_groups):
        n_sites = 0
        for j in range(n_column):
            if np.isnan(gr_start[i, j]):
                continue
            valid_ind = np.logical_and(
                plot_sites_column[j] >= gr_start[i, j],
                plot_sites_column[j] < gr_end[i, j],
            )
            n_sites += np.sum(valid_ind)
        region_site_num.append(n_sites)

    return region_site_num


def get_label_name(label_info, region_label):
    label_names = []
    label_acronym = []
    label_color = []
    for i in range(len(region_label)):
        if region_label[i] == 0:
            label_names.append(" ")
            label_acronym.append(" ")
            label_color.append((128, 128, 128))
        else:
            da_ind = np.where(np.ravel(label_info["index"]) == region_label[i])[0][0]
            label_names.append(label_info["label"][da_ind])
            label_acronym.append(label_info["abbrev"][da_ind])
            label_color.append(label_info["color"][da_ind])

    label_color = np.asarray(label_color)
    return label_names, label_acronym, label_color


def get_sites_loc_related_to_base_center(probe_settings, probe_length_without_tip_um):
    probe_type_name = probe_settings["probe_type_name"]
    probe_thickness_um = probe_settings["probe_thickness"]
    per_max_sites = probe_settings["per_max_sites"]
    sites_distance_um = probe_settings["sites_distance"]
    x_bias_um = probe_settings["x_bias"]
    y_bias_um = probe_settings["y_bias"]

    if probe_type_name == "Tetrode":
        sites_loc = [np.array([[0, 0, 0]]) for _ in range(4)]
    else:
        sites_loc = []  # inverse r-vals, u-vals, n-vals
        n_val = 0.5 * probe_thickness_um
        for i in range(len(x_bias_um)):
            possible_n_sites = int(
                (probe_length_without_tip_um - y_bias_um[i]) / sites_distance_um[i]
            )
            if possible_n_sites <= per_max_sites[i]:
                r_vals = (
                    np.arange(possible_n_sites) * sites_distance_um[i] + y_bias_um[i]
                )
            else:
                r_vals = (
                    np.arange(per_max_sites[i]) * sites_distance_um[i] + y_bias_um[i]
                )
            num_sites = len(r_vals)
            u_vals = np.repeat(x_bias_um[i], num_sites)
            n_vals = np.repeat(n_val, num_sites)
            temp = np.stack([r_vals, u_vals, n_vals], axis=1)
            sites_loc.append(temp)
    return sites_loc


def merge_sites_into_line(sites_loc):
    y_vec = []
    for i in range(len(sites_loc)):
        y_vec.append(sites_loc[i][:, 0])
    y_vec = np.concatenate(y_vec)
    y_vals = np.unique(y_vec)
    sites_loc_line = []
    sites_count = []
    temp = []
    for i in range(len(y_vals)):
        n = len(np.where(y_vec == y_vals[i])[0])
        temp.append(np.array([y_vals[i], 0, 0]))
        sites_count.append(n)
    temp = np.asarray(temp)
    sites_loc_line.append(temp)

    return sites_loc_line, sites_count


def get_pnt_from_loc(sct, loc, n_vec, u_vec, r_vec, bregma, vox_size):
    pnt = []
    pnt_vox = []
    for i in range(len(loc)):
        temp = [
            sct * vox_size
            + r_vec * loc[i][ind, 0]
            + u_vec * loc[i][ind, 1]
            + n_vec * loc[i][ind, 2]
            for ind in range(len(loc[i]))
        ]
        temp = np.asarray(temp)
        temp = temp / vox_size
        pnt.append(temp)
        vox_temp = temp + bregma
        vox_temp = vox_temp.astype(int)
        pnt_vox.append(vox_temp)

    return pnt, pnt_vox


def get_column_loc(
    sites_loc_to_base, probe_length_without_tip_um, vxsize_um, verbose=False
):
    sites_r_vals = []
    for i in range(len(sites_loc_to_base)):
        sites_r_vals.append(sites_loc_to_base[i][:, 0])
    sites_r_vals = np.concatenate(sites_r_vals)
    uni_sites_r_vals = np.unique(sites_r_vals)

    if len(uni_sites_r_vals) > 1:
        if uni_sites_r_vals[0] == 0:
            vis_r_vals = np.arange(uni_sites_r_vals[0], uni_sites_r_vals[1], vxsize_um)
            for i in range(1, len(uni_sites_r_vals) - 1):
                vis_r_vals = np.append(
                    vis_r_vals,
                    np.arange(uni_sites_r_vals[i], uni_sites_r_vals[i + 1], vxsize_um),
                )
        else:
            vis_r_vals = np.arange(0, uni_sites_r_vals[0], vxsize_um)
            for i in range(len(uni_sites_r_vals) - 1):
                vis_r_vals = np.append(
                    vis_r_vals,
                    np.arange(uni_sites_r_vals[i], uni_sites_r_vals[i + 1], vxsize_um),
                )

        if uni_sites_r_vals[-1] != probe_length_without_tip_um:
            vis_r_vals = np.append(
                vis_r_vals,
                np.arange(uni_sites_r_vals[-1], probe_length_without_tip_um, vxsize_um),
            )
        vis_r_vals = np.append(vis_r_vals, probe_length_without_tip_um)
    else:
        vis_r_vals = np.arange(0, probe_length_without_tip_um, vxsize_um)

    column_loc = []
    for i in range(len(sites_loc_to_base)):
        temp = np.zeros((len(vis_r_vals), 3))
        temp[:, 0] = vis_r_vals
        temp[:, 1] = sites_loc_to_base[i][0, 1]
        temp[:, 2] = sites_loc_to_base[i][0, 2]
        column_loc.append(temp)

    if verbose:
        print("uni_sites_r_vals")
        print(uni_sites_r_vals)

        print("vis_r_vals")
        print(vis_r_vals)

        print("column_loc")
        print(column_loc)

    return column_loc


def get_loc_related_to_start(loc_to_base, probe_length_without_tip_um):
    loc_to_start = []  # inverse r-vals, u-vals, n-vals
    for i in range(len(loc_to_base)):
        temp_loc = loc_to_base[i].copy()
        temp_loc[:, 0] = probe_length_without_tip_um - temp_loc[:, 0]
        loc_to_start.append(temp_loc)
    return loc_to_start


def get_fine_label_matrix(column_vox, label_data, verbose=False):
    fine_label_mat = []
    for i in range(len(column_vox)):
        fine_label_mat.append(
            label_data[column_vox[i][:, 0], column_vox[i][:, 1], column_vox[i][:, 2]]
        )
    fine_label_mat = np.asarray(fine_label_mat).T

    if verbose:
        # print out fine label matrix
        indexs = np.arange(0, len(fine_label_mat), 20)
        for i in range(len(indexs) - 1):
            print(fine_label_mat[indexs[i] : indexs[i + 1]])
        print(fine_label_mat[indexs[-1] : len(fine_label_mat)])

    return fine_label_mat


def calculate_probe_info(
    data_list,
    pieces_names,
    label_data,
    label_info,
    vxsize_um,
    probe_settings,
    merge_sites,
    bregma,
    site_face,
    n_hat,
    pre_plan,
):
    """

    :param data: 3d coordinates for all the points
    :param label_data: original brain region segmentation
    :param label_info:
    :param vxsize_um:
    :param tip_length:
    :param channel_size:
    :param bregma:
    :return:
    """

    data_dict = None
    data = data_list[0]
    for i in range(1, len(data_list)):
        data = np.vstack([data, data_list[i]])
    # print('data', data)

    # start_pnt and end_pnt are coordinates related to the given Bregma
    probe_type_name = probe_settings["probe_type_name"]
    probe_max_length_um = probe_settings["probe_length"]
    tip_length_um = probe_settings["tip_length"]

    # get direction and probe center start and end (pc - probe center)
    pc_start_pnt, pc_end_pnt, avg, direction = line_fit(data)
    direction = pc_end_pnt - pc_start_pnt
    direction = direction / np.linalg.norm(direction)
    pc_start_vox = pc_start_pnt + bregma
    pc_end_vox = pc_end_pnt + bregma

    # print('direction', direction)

    # get angels
    ap_angle, ml_angle = get_angles(direction)
    # print('ap_angle', ap_angle)
    # print(ml_angle)

    # print('check-start_pnt', pc_start_pnt)
    # # print(pc_start_vox)
    # correct probe center start point
    pc_sp, error_index = correct_start_pnt(
        label_data, pc_start_pnt, pc_start_vox, direction, bregma, verbose=True
    )
    if error_index != 0:
        return data_dict, 15

    pc_ep, probe_length_with_tip_um, probe_length_without_tip_um = correct_end_point(
        pc_sp,
        pc_end_pnt,
        direction,
        vxsize_um,
        tip_length_um,
        probe_max_length_um,
        verbose=True,
    )

    pv_sp = pc_sp + bregma
    pv_sp = pv_sp.astype(int)
    pv_ep = pc_ep + bregma
    pv_ep = pv_ep.astype(int)

    enter_coords = pc_sp * vxsize_um
    end_coords = pc_ep * vxsize_um

    # print('')

    dv = (pc_sp[2] - pc_ep[2]) * vxsize_um
    ap_tilt, ml_tilt = get_tilt_info(pc_sp, pc_ep)

    if pre_plan:
        r_hat = direction.copy()
        u_hat = np.cross(n_hat, r_hat)

        n_vec, u_vec = get_vector_according_to_site_face(n_hat, u_hat, site_face)
    else:
        r_hat, u_vec, n_vec = calculate_vector_according_to_site_face(
            direction, site_face
        )

    # sites location, list of  [(n_sites, 3),...], in um
    sites_loc_to_base_temp = get_sites_loc_related_to_base_center(
        probe_settings, probe_length_without_tip_um
    )
    # print('sites_loc_to_base')
    # print(sites_loc_to_base)

    if merge_sites or probe_type_name == "Tetrode":
        sites_loc_to_base, sites_line_count = merge_sites_into_line(
            sites_loc_to_base_temp
        )
    else:
        sites_loc_to_base = sites_loc_to_base_temp.copy()
        sites_line_count = None

    # column location list of [(n_locs, 3), ...], in um
    column_loc_to_base = get_column_loc(
        sites_loc_to_base, probe_length_without_tip_um, vxsize_um, verbose=False
    )

    # column location relative to the start, from bottom to top, in um
    column_loc = get_loc_related_to_start(
        column_loc_to_base, probe_length_without_tip_um
    )
    # print('column_loc')
    # print(column_loc)

    # column points (related to bregma) and column vox
    column_pnt, column_vox = get_pnt_from_loc(
        pc_sp, column_loc, n_vec, u_vec, r_hat, bregma, vxsize_um
    )
    # print('column_pnt')
    # print(column_pnt)
    # print('column_vox')
    # print(column_vox)

    fine_label_mat = get_fine_label_matrix(column_vox, label_data, verbose=False)
    group_mat, group_id_label = group_labels(fine_label_mat)
    label_names, label_acronym, label_color = get_label_name(label_info, group_id_label)

    if probe_type_name != "Tetrode":
        # sites loc related to the start, in um
        sites_loc = get_loc_related_to_start(
            sites_loc_to_base, probe_length_without_tip_um
        )
        # print(sites_loc)

        sites_pnt, sites_vox = get_pnt_from_loc(
            pc_sp, sites_loc, n_vec, u_vec, r_hat, bregma, vxsize_um
        )
        # print(sites_pnt)
        # print('sites_vox')
        # print(sites_vox)
    else:
        sites_pnt = [np.array([pc_ep])]
        sites_vox_temp = sites_pnt[0] + bregma
        sites_vox = [sites_vox_temp.astype(int)]

    sites_label = []
    for i in range(len(sites_vox)):
        sites_label.append(
            label_data[sites_vox[i][:, 0], sites_vox[i][:, 1], sites_vox[i][:, 2]]
        )

    # print('sites_label')
    # print(sites_label)

    vis_data, region_length, region_site_num, region_text_loc = get_vis_data(
        group_mat, column_loc_to_base, sites_loc_to_base, sites_line_count, vxsize_um
    )

    data_dict = {
        "object_name": "probe",
        "probe_type_name": probe_type_name,
        "data": data_list,
        "pieces_names": pieces_names,
        "ap_tilt": ap_tilt,
        "ml_tilt": ml_tilt,
        "insertion_coords_3d": pc_sp,
        "terminus_coords_3d": pc_ep,
        "direction": direction,
        "probe_length": probe_length_with_tip_um,
        "dv": dv,
        "ap_angle": ap_angle,
        "ml_angle": ml_angle,
        "insertion_coords": enter_coords,
        "insertion_vox": pv_sp,
        "terminus_coords": end_coords,
        "terminus_vox": pv_ep,
        "sites_label": sites_label,
        "sites_loc_b": sites_pnt,
        "sites_vox": sites_vox,
        "region_label": group_id_label,
        "region_length": region_length,
        "region_sites": region_site_num,
        "text_loc": region_text_loc,
        "label_name": label_names,
        "label_acronym": label_acronym,
        "label_color": label_color,
        "vis_data": vis_data,
    }
    return data_dict, error_index


def get_pre_multi_shank_vis_base(x_vals, y_vals):
    x_vals = np.ravel(x_vals)
    y_vals = np.ravel(y_vals)
    valid_ind = np.where(y_vals == 0)[0]
    if len(valid_ind) != 0:
        base_loc = x_vals[valid_ind]
    else:
        base_loc = np.array([0])
    # if site_face in [0, 1]:
    #     valid_ind = np.where(y_vals == 0)[0]
    #     if len(valid_ind) != 0:
    #         base_loc = x_vals[valid_ind]
    #     else:
    #         base_loc = np.array([0])
    # else:
    #     valid_ind = np.where(x_vals == 0)[0]
    #     if len(valid_ind) != 0:
    #         base_loc = y_vals[valid_ind]
    #     else:
    #         base_loc = np.array([0])
    return base_loc


def get_vector_according_to_site_face(n_hat, u_hat, site_face):
    # for pre-plan - 2d plan
    if site_face == 0:
        # site face out, facing to you
        n_vec = n_hat.copy()
        u_vec = u_hat.copy()
    elif site_face == 1:
        # site face in, facing away from you
        n_vec = -n_hat
        u_vec = -u_hat
    elif site_face == 2:
        # site face left, facing to you left-hand side
        n_vec = -u_hat
        u_vec = n_hat.copy()
    elif site_face == 3:
        # site face right, facing to you right-hand side
        n_vec = u_hat.copy()
        u_vec = -n_hat
    else:
        raise ValueError("no such site face, stupid!!!!!!")

    # print('n_vec', n_vec)
    # print('u_vec', u_vec)

    return n_vec, u_vec


def calculate_vector_according_to_site_face(direction, site_face):
    # for after surgery
    r_hat = direction.copy()
    if check_parallel_to_z(direction):
        if site_face == 0:
            n_hat = np.array([0, 1, 0])
            u_hat = np.cross(n_hat, r_hat)
        elif site_face == 1:
            n_hat = np.array([0, -1, 0])
            u_hat = np.cross(n_hat, r_hat)
        elif site_face == 2:
            n_hat = np.array([-1, 0, 0])
            u_hat = np.cross(n_hat, r_hat)
        elif site_face == 3:
            n_hat = np.array([1, 0, 0])
            u_hat = np.cross(n_hat, r_hat)
        else:
            n_hat = None
            u_hat = None
            print("Site face can only be 0-Up, 1-Down, 2-Left, 3-Right.")
    else:
        if site_face == 0:
            t_hat = np.array([-r_hat[1], r_hat[0], 0])
            u_hat = t_hat / np.linalg.norm(t_hat)
            n_hat = np.cross(r_hat, u_hat)
        elif site_face == 1:
            t_hat = np.array([r_hat[1], r_hat[0], 0])
            u_hat = t_hat / np.linalg.norm(t_hat)
            n_hat = np.cross(r_hat, u_hat)
        elif site_face == 2:
            t_hat = np.array([-r_hat[1], r_hat[0], 0])
            n_hat = t_hat / np.linalg.norm(t_hat)
            u_hat = np.cross(n_hat, r_hat)
        elif site_face == 3:
            t_hat = np.array([r_hat[1], r_hat[0], 0])
            n_hat = t_hat / np.linalg.norm(t_hat)
            u_hat = np.cross(n_hat, r_hat)
        else:
            n_hat = None
            u_hat = None
            print("Site face can only be 0-Up, 1-Down, 2-Left, 3-Right.")
        # print('t_hat', t_hat)
        # print(n_hat)
        # print(u_hat)
    return r_hat, u_hat, n_hat


def get_center_lines(
    pnts, r_hat, n_hat, u_hat, x_vals, y_vals, length, site_face, vox_size
):
    n_vec, u_vec = get_vector_according_to_site_face(n_hat, u_hat, site_face)

    # print(x_vals)
    # print(y_vals)

    line_data = []
    for i in range(len(x_vals)):
        s_val = u_vec * x_vals[i] + n_vec * y_vals[i]
        # print('s_val', s_val)
        temp = [s_val, s_val + r_hat * length * vox_size]
        # print(temp)
        line_data.append(pnts[0] + np.asarray(temp) / vox_size)

    return line_data


def get_pre_ms_vis_base(multi_shanks, site_face, atlas_display):
    if multi_shanks is None:
        base_loc = np.array([0])
    else:
        if atlas_display == "sagittal":
            if site_face not in [0, 1]:
                base_loc = np.ravel(multi_shanks)
            else:
                base_loc = np.array([0])
        else:
            if site_face in [0, 1]:
                base_loc = np.ravel(multi_shanks)
            else:
                base_loc = np.array([0])
    return base_loc


def get_pre_mp_vis_base(base_loc, atlas_display):
    if atlas_display == "sagittal":
        valid_ind = np.where(base_loc[:, 1] == 0)[0]
        if len(valid_ind) == 0:
            vis_base = np.array([0])
        else:
            vis_base = base_loc[valid_ind, 1]
    else:
        valid_ind = np.where(base_loc[:, 0] == 0)[0]
        if len(valid_ind) == 0:
            vis_base = np.array([0])
        else:
            vis_base = base_loc[valid_ind, 0]
    return vis_base


class Probe(object):
    def __init__(self):
        self.probe_length = None
        self.tip_length = None
        self.site_width = None
        self.site_height = None
        self.total_sites = None
        self.site_number_in_banks = None
        self.sites_distance = None
        self.probe_type = None
        self.x_bias = None
        self.per_max_sites = None
        self.sites_distance = None
        self.y_bias = None
        self.probe_type_name = None
        self.probe_thickness = None
        self.multi_shanks = None
        self.exist_probes = None
        self.faces = None

        self.set_np1()

    def extend_exist_probes(self, prb):
        pass

    def get_exist_probes(self):
        self.exist_probes = {"NP1.0": None, "NP2.0": None, "Tetrode": None}

    def set_np2(self):
        self.probe_type = 1
        self.probe_type_name = "NP2.0"
        self.probe_thickness = 24
        self.probe_length = 10000
        self.tip_length = 175
        self.site_width = 16
        self.site_height = 15
        self.x_bias = [-8, 16]
        self.per_max_sites = [480, 480]
        self.sites_distance = [15, 15]
        self.y_bias = [7.5, 7.5]
        self.site_number_in_banks = (384, 384, 192)
        self.multi_shanks = [-375, -125, 125, 375]
        self.faces = "Front"

    def set_np1(self):
        self.probe_type = 0
        self.probe_type_name = "NP1.0"
        self.probe_thickness = 24
        self.probe_length = 10000
        self.tip_length = 175
        self.site_width = 16
        self.site_height = 20
        self.x_bias = [-16, -8, 8, 16]
        self.per_max_sites = [240, 240, 240, 240]
        self.sites_distance = [40, 40, 40, 40]
        self.y_bias = [30, 10, 30, 10]
        self.site_number_in_banks = (384, 384, 192)
        self.multi_shanks = None
        self.faces = "Front"

    def set_tetrode(self):
        self.probe_type = 3
        self.probe_type_name = "Tetrode"
        self.probe_thickness = 0
        self.probe_length = None
        self.tip_length = 0
        self.site_width = None
        self.site_height = None
        self.x_bias = 0
        self.per_max_sites = 4
        self.sites_distance = 0
        self.y_bias = 0
        self.site_number_in_banks = None
        self.multi_shanks = None
        self.faces = None

    def set_linear_silicon(self, pss):
        self.probe_type = 2
        self.probe_type_name = "Linear-Silicon"
        self.probe_length = pss["probe_length"]
        self.probe_thickness = pss["probe_thickness"]
        self.tip_length = pss["tip_length"]
        self.site_width = pss["site_width"]
        self.site_height = pss["site_height"]

        self.per_max_sites = pss["per_max_sites"]
        self.sites_distance = pss["sites_distance"]
        self.x_bias = pss["x_bias"]
        self.y_bias = pss["y_bias"]
        self.site_number_in_banks = None
        self.multi_shanks = None
        self.faces = "Front"

    def get_settings(self):
        data = {
            "probe_type": self.probe_type,
            "probe_type_name": self.probe_type_name,
            "probe_thickness": self.probe_thickness,
            "probe_length": self.probe_length,
            "tip_length": self.tip_length,
            "site_height": self.site_height,
            "site_width": self.site_width,
            "per_max_sites": self.per_max_sites,
            "sites_distance": self.sites_distance,
            "x_bias": self.x_bias,
            "y_bias": self.y_bias,
            "site_number_in_banks": self.site_number_in_banks,
            "multi_shanks": self.multi_shanks,
        }
        return data

    def probe_faces_changed(self, face_direction):
        self.faces = face_direction

    def get_multi_shank_3d_base(self):
        if self.multi_shanks is None:
            return
        if self.faces in ["Front", "Back"]:
            points3 = np.zeros((len(self.multi_shanks), 3))
            points3[:, 1] = self.multi_shanks
        else:
            points3 = np.zeros((len(self.multi_shanks), 3))
            points3[:, 2] = self.multi_shanks
        return points3


class MultiProbes(object):
    def __init__(self):
        self.x_vals = None
        self.y_vals = None
        self.faces = None

    def set_multi_probes(self, multi_settings):
        self.x_vals = multi_settings["x_vals"]
        self.y_vals = multi_settings["y_vals"]
        self.faces = multi_settings["faces"]

    def get_multi_settings(self):
        if self.x_vals is None:
            multi_settings = None
        else:
            if not isinstance(self.x_vals, list):
                x_vals = self.x_vals.tolist()
            else:
                x_vals = self.x_vals.copy()

            if not isinstance(self.y_vals, list):
                y_vals = self.y_vals.tolist()
            else:
                y_vals = self.y_vals.copy()

            if not isinstance(self.faces, list):
                faces = self.faces.tolist()
            else:
                faces = self.faces.copy()
            multi_settings = {"x_vals": x_vals, "y_vals": y_vals, "faces": faces}
        return multi_settings

    def check_multi_settings(self):
        x_unique = np.unique(self.x_vals)

        for x_val in x_unique:
            v_ind = np.where(np.ravel(self.x_vals) == x_val)[0]
            if len(v_ind) > 1:
                if len(np.unique(np.ravel(self.y_vals)[v_ind])) != len(v_ind):
                    msg = "There are at least 2 probes located at the same location."
                    return msg

        return

    # def get_base_loc_3d(self, multi_shank):
    #     if multi_shank is None:
    #         points3 = np.zeros((len(self.x_vals), 3))
    #         points3[:, 1] = self.x_vals
    #         points3[:, 2] = self.y_vals
    #     else:
    #         n_shank = len(multi_shank)
    #         points3 = []
    #         if self.faces[0] in ['Top', 'Bottom']:
    #             for i in range(len(self.x_vals)):
    #                 p3 = np.zeros((n_shank, 3))
    #                 p3[:, 1] = multi_shank + self.x_vals[i]
    #                 p3[:, 2] = self.y_vals[i]
    #                 points3.append(p3)
    #         else:
    #             for i in range(len(self.x_vals)):
    #                 p3 = np.zeros((n_shank, 3))
    #                 p3[:, 1] = self.x_vals[i]
    #                 p3[:, 2] = multi_shank + self.y_vals[i]
    #                 points3.append(p3)
    #         points3 = np.concatenate(points3)
    #     return points3
    #
    # def get_base_loc_2d(self, multi_shank):
    #     if multi_shank is None:
    #         points2 = np.zeros((len(self.x_vals), 2))
    #         points2[:, 0] = self.x_vals
    #         points2[:, 1] = self.y_vals
    #     else:
    #         n_shank = len(multi_shank)
    #         points2 = []
    #         if self.faces[0] in ['Top', 'Bottom']:
    #             for i in range(len(self.x_vals)):
    #                 p2 = np.zeros((n_shank, 2))
    #                 p2[:, 0] = multi_shank + self.x_vals[i]
    #                 p2[:, 1] = self.y_vals[i]
    #                 points2.append(p2)
    #         else:
    #             for i in range(len(self.x_vals)):
    #                 p2 = np.zeros((n_shank, 2))
    #                 p2[:, 0] = self.x_vals[i]
    #                 p2[:, 1] = multi_shank + self.y_vals[i]
    #                 points2.append(p2)
    #         points2 = np.concatenate(points2)
    #     return points2



import os


import pickle
import nrrd
import csv
import nibabel as nib
import numpy as np
import pandas as pd
import cv2
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
import pyqtgraph.opengl as gl

from .uuuuuu import read_qss_file, make_contour_img, read_excel_file, hex2rgb
from .obj_items import render_volume, render_small_volume
from .atlas_loader import process_atlas_raw_data, AtlasLoader, check_data_path_and_load


class CustomerAtlasWorker(QObject):
    finished = pyqtSignal()
    progress = pyqtSignal(float)
    error_occur = pyqtSignal(str)

    def __init__(self):
        super(CustomerAtlasWorker, self).__init__()
        self.saving_folder = None
        self.data_local = None
        self.segmentation_local = None
        self.label_local = None
        self.mask_local = None
        self.b_val = None
        self.l_val = None
        self.vox_size = None
        self.factor = 2
        self.axis_info = None

    def set_data(
        self,
        saving_folder,
        data_local,
        segmentation_local,
        label_local,
        axis_info,
        b_val,
        vox_size,
        mask_local=None,
        factor=2,
    ):
        self.saving_folder = saving_folder
        self.data_local = data_local
        self.segmentation_local = segmentation_local
        self.label_local = label_local
        self.mask_local = mask_local
        self.b_val = b_val
        self.vox_size = vox_size
        self.axis_info = axis_info
        self.factor = factor

    def progress_control(self, total_count):
        self.progress.emit(total_count)

    def run(self):
        self.progress.emit(1)
        if self.vox_size < 1e-4:
            self.error_occur.emit("Please set voxel size.")
            return
        print(self.saving_folder)
        print(self.label_local)
        df, msg = read_excel_file(os.path.join(self.saving_folder, self.label_local))
        if msg is not None:
            self.error_occur.emit(msg)
            return

        reformat_keys = list(df.columns)
        for i in range(len(reformat_keys)):
            reformat_keys[i] = reformat_keys[i].lower()
        df.columns = reformat_keys
        self.progress.emit(2)
        try:
            da_labels = df["name"].values
            self.progress.emit(3)

            da_short_label = df["acronym"].values

        except KeyError:
            self.error_occur.emit('Label file missing columns "name" or "acronym".')
            return
        self.progress.emit(4)

        try:
            levels = []
            structure_id_path = df["structure_id_path"].values
            for i in range(len(structure_id_path)):
                da_path = structure_id_path[i]
                da_path_split = da_path.split("/")
                for j in np.arange(len(da_path_split))[::-1]:
                    if da_path_split[j] == "":
                        da_path_split.pop(j)
                levels.append(len(da_path_split))
        except KeyError:
            self.error_occur.emit('Label file missing columns "structure_id_path".')
            return
        self.progress.emit(5)

        try:
            hex_colors = df["color_hex_triplet"].values
            rgb_colors = []
            for i in range(len(hex_colors)):
                r, g, b = hex2rgb(hex_colors[i])
                rgb_colors.append([r, g, b])
            rgb_colors = np.asarray(rgb_colors)
        except KeyError:
            rgb_colors = []
            for i in range(len(da_short_label)):
                r, g, b = np.random.randint(0, 255, 3)
                rgb_colors.append([r, g, b])
            rgb_colors = np.asarray(rgb_colors)
        self.progress.emit(6)

        try:
            parent = df["parent_id"].values.astype(int)
            ids = df["id"].values.astype(int)
        except KeyError:
            self.error_occur.emit(
                'Label file missing columns "parent_structure_id" or "id".'
            )
            return

        label_info = {
            "index": ids,
            "label": da_labels,
            "parent": parent,
            "abbrev": da_short_label,
            "color": rgb_colors,
            "level_indicator": levels,
        }

        with open(os.path.join(self.saving_folder, "atlas_labels.pkl"), "wb") as handle:
            pickle.dump(label_info, handle, protocol=pickle.HIGHEST_PROTOCOL)
        self.progress.emit(9)

        # laod atlas
        atlas_data, success = check_data_path_and_load(
            os.path.join(self.saving_folder, self.data_local)
        )
        if not success:
            self.error_occur.emit(
                "Failed to load atlas data file. Currently only support for .nii and .nrrd."
            )
            return
        atlas_data = atlas_data - np.min(atlas_data)
        atlas_size = atlas_data.shape
        if np.any(np.ravel(atlas_size) < self.factor):
            self.error_occur.emit("Factor can not be larger than atlas size.")
        self.progress.emit(14)
        # laod segmentation data
        segmentation_data, success = check_data_path_and_load(
            os.path.join(self.saving_folder, self.segmentation_local)
        )
        if not success:
            self.error_occur.emit(
                "Failed to load segmentation data file. Currently only support for .nii and .nrrd."
            )
            return
        self.progress.emit(19)
        if not (segmentation_data.shape == atlas_data.shape):
            self.error_occur.emit("Atlas shape is different than segmentation shape.")
            return

        if self.mask_local is not None:
            mask_data, success = check_data_path_and_load(
                os.path.join(self.saving_folder, self.mask_local)
            )
            if not success:
                self.error_occur.emit(
                    "Failed to load mask data file. Currently only support for .nii and .nrrd."
                )
                return
            self.progress.emit(23)
            if not mask_data.shape == atlas_data.shape:
                self.error_occur.emit("Atlas shape is different than mask shape.")
                return

            # make segmentation with mask
            progress_step = np.linspace(23, 26, len(mask_data))
            for i in range(len(mask_data)):
                self.progress.emit(progress_step[i])
                segmentation_data[i][mask_data[i] == 0] = 0
            segmentation_data = segmentation_data.astype("int")

            progress_step = np.linspace(26, 28, len(mask_data))
            for i in range(len(mask_data)):
                self.progress.emit(progress_step[i])
                atlas_data[i][mask_data[i] == 0] = 0

        atlas_data = atlas_data / np.max(atlas_data)
        self.progress.emit(30)

        unique_label = np.unique(segmentation_data)
        n_unique_labels = len(unique_label)
        self.progress.emit(35)

        if self.axis_info["direction_change"][0]:
            segmentation_data = segmentation_data[::-1, :, :]
            atlas_data = atlas_data[::-1, :, :]
        self.progress.emit(36)
        if self.axis_info["direction_change"][1]:
            segmentation_data = segmentation_data[:, ::-1, :]
            atlas_data = atlas_data[:, ::-1, :]
        self.progress.emit(37)
        if self.axis_info["direction_change"][2]:
            segmentation_data = segmentation_data[:, :, ::-1]
            atlas_data = atlas_data[:, :, ::-1]
        self.progress.emit(38)

        segmentation_data = np.transpose(segmentation_data, self.axis_info["to_HERBS"])
        segmentation_data = segmentation_data.astype(int)
        # print(segmentation_data.shape)
        self.progress.emit(39)

        segment = {"data": segmentation_data, "unique_label": unique_label}

        outfile = open(os.path.join(self.saving_folder, "segment_pre_made.pkl"), "wb")
        pickle.dump(segment, outfile)
        outfile.close()
        self.progress.emit(42)

        new_atlas_shape = atlas_data.shape
        b_val = np.ravel(self.b_val)[np.array(self.axis_info["to_HERBS"])]
        if b_val[0] == 0:
            b_val[0] = int(new_atlas_shape[0] / 2)
        self.progress.emit(43)
        if b_val[1] == 0:
            b_val[1] = int(new_atlas_shape[1] / 2)
        self.progress.emit(44)

        for i in range(3):
            if self.axis_info["direction_change"][i]:
                b_val[i] = new_atlas_shape[i] - 1 - b_val[i]

        if b_val[2] == 0:
            b_val[2] = new_atlas_shape[2] - 1 - b_val[2]

        atlas_info = [
            {
                "name": "anterior",
                "values": np.arange(atlas_data.shape[0]) * self.vox_size,
                "units": "um",
            },
            {
                "name": "dorsal",
                "values": np.arange(atlas_data.shape[1]) * self.vox_size,
                "units": "um",
            },
            {
                "name": "right",
                "values": np.arange(atlas_data.shape[2]) * self.vox_size,
                "units": "um",
            },
            {"vxsize": self.vox_size, "Bregma": [b_val[0], b_val[1], b_val[2]]},
        ]
        self.progress.emit(45)
        atlas = {"data": atlas_data, "info": atlas_info}

        outfile = open(os.path.join(self.saving_folder, "atlas_pre_made.pkl"), "wb")
        pickle.dump(atlas, outfile)
        outfile.close()
        self.progress.emit(50)

        mesh_data = render_volume(
            atlas_data, self.saving_folder, factor=self.factor, level=0.1
        )
        self.progress.emit(55)

        mesh_path = os.path.join(self.saving_folder, "meshes")
        if not os.path.exists(mesh_path):
            os.mkdir(mesh_path)

        progress_step = np.linspace(55, 68, n_unique_labels)
        for i in range(n_unique_labels):
            self.progress.emit(progress_step[i])
            label_id = int(unique_label[i])

            if label_id == 0:
                continue
            render_small_volume(
                label_id,
                mesh_path,
                atlas_data,
                segmentation_data,
                factor=self.factor,
                level=0.1,
            )

        small_mesh_list = {}
        file_list = os.listdir(mesh_path)
        for da_file in file_list:
            file_name = os.path.basename(da_file)
            da_name, file_extension = os.path.splitext(file_name)
            if file_extension == ".pkl":
                infile = open(os.path.join(mesh_path, da_file), "rb")
                md = pickle.load(infile)
                infile.close()

                small_mesh_list[str(da_name)] = md

        self.progress.emit(69)
        outfile = open(
            os.path.join(self.saving_folder, "atlas_small_meshdata.pkl"), "wb"
        )
        pickle.dump(small_mesh_list, outfile)
        outfile.close()
        self.progress.emit(70)

        segment_data_shape = segmentation_data.shape

        sagital_contour_img = np.zeros(segment_data_shape, "i")
        coronal_contour_img = np.zeros(segment_data_shape, "i")
        horizontal_contour_img = np.zeros(segment_data_shape, "i")

        # pre-process boundary ----- todo: change this part as optional
        process_index = np.linspace(70, 78, segment_data_shape[0])
        for i in range(segment_data_shape[0]):
            self.progress.emit(process_index[i])
            da_slice = segmentation_data[i, :, :].copy()
            contour_img = make_contour_img(da_slice)
            sagital_contour_img[i, :, :] = contour_img

        outfile_ct = open(
            os.path.join(self.saving_folder, "sagital_contour_pre_made.pkl"), "wb"
        )
        pickle.dump(sagital_contour_img, outfile_ct)
        outfile_ct.close()
        self.progress.emit(80)

        process_index = np.linspace(80, 88, segment_data_shape[1])
        for i in range(segment_data_shape[1]):
            self.progress.emit(process_index[i])
            da_slice = segmentation_data[:, i, :].copy()
            contour_img = make_contour_img(da_slice)
            coronal_contour_img[:, i, :] = contour_img

        outfile_ct = open(
            os.path.join(self.saving_folder, "coronal_contour_pre_made.pkl"), "wb"
        )
        pickle.dump(coronal_contour_img, outfile_ct)
        outfile_ct.close()
        self.progress.emit(90)

        process_index = np.linspace(90, 98, segment_data_shape[2])
        for i in range(segment_data_shape[2]):
            self.progress.emit(process_index[i])
            da_slice = segmentation_data[:, :, i].copy()
            contour_img = make_contour_img(da_slice)
            horizontal_contour_img[:, :, i] = contour_img

        outfile_ct = open(
            os.path.join(self.saving_folder, "horizontal_contour_pre_made.pkl"), "wb"
        )
        pickle.dump(horizontal_contour_img, outfile_ct)
        outfile_ct.close()

        # saving atlas axis changing information
        self.axis_info["size"] = tuple(atlas_size)
        outfile_axis = open(
            os.path.join(self.saving_folder, "atlas_axis_info.pkl"), "wb"
        )
        pickle.dump(self.axis_info, outfile_axis)
        outfile_axis.close()

        self.progress.emit(100)

        self.finished.emit()


class AtlasProcessor(QDialog):
    def __init__(self):
        super().__init__()
        qss_file_name = "qss/dialogs.qss"
        qss_style_sheet = read_qss_file(qss_file_name)
        self.setStyleSheet(qss_style_sheet)
        self.setWindowTitle("Atlas Processor")

        self.folder_path = None
        self.data_local = None
        self.segmentation_local = None
        self.mask_local = None
        self.label_local = None
        self.bregma_coord = [0, 0, 0]
        self.lambda_coord = [0, 0, 0]
        self.voxel_size = 0
        self.factor_val = 2
        self.axis_info = None
        self.directions = [0, 0, 0]
        self.dim_group = [-1, -1, -1]
        self.info_flag = True

        self.thread = None
        self.worker = None

        self.selector_vals = [
            "L.H. --> R.H.",
            "R.H. --> L.H.",
            "Post. --> Ant.",
            "Ant. --> Post.",
            "Sup. --> Inf.",
            "Inf. --> Sup.",
        ]
        self.group_maps = np.array([1, 1, 2, 2, 3, 3])

        layout = QGridLayout(self)

        box_label_style = read_qss_file("qss/box_label.qss")
        data_label = QLabel("Volume File:")
        self.data_btn = QPushButton("Select File")
        self.data_btn.setAutoDefault(False)
        self.data_btn.setFocus(False)
        self.data_line = QLabel()
        self.data_line.setStyleSheet(box_label_style)

        seg_label = QLabel("Segmentation File: ")
        self.seg_btn = QPushButton("Select File")
        self.seg_btn.setAutoDefault(False)
        self.seg_btn.setFocus(False)
        self.seg_line = QLabel()
        self.seg_line.setStyleSheet(box_label_style)

        mask_label = QLabel("Mask File (optional): ")
        self.mask_btn = QPushButton("Select File")
        self.mask_btn.setAutoDefault(False)
        self.mask_btn.setFocus(False)
        self.mask_line = QLabel()
        self.mask_line.setStyleSheet(box_label_style)

        labinf_label = QLabel("Label Information File:")
        self.labinf_btn = QPushButton("Select File")
        self.labinf_btn.setAutoDefault(False)
        self.labinf_btn.setFocus(False)
        self.labinf_line = QLabel()
        self.labinf_line.setStyleSheet(box_label_style)

        valid_input = QIntValidator(0, 99999)

        float_input = QRegExpValidator(QRegExp(r"[0-9].+"))
        # float_input = QDoubleValidator(0.0, 100.0, 6)

        bregma_label = QLabel("Bregma Coordinates (voxel): ")
        self.bregma_input1 = QLineEdit("0")
        self.bregma_input1.setValidator(valid_input)
        self.bregma_input2 = QLineEdit("0")
        self.bregma_input2.setValidator(valid_input)
        self.bregma_input3 = QLineEdit("0")
        self.bregma_input3.setValidator(valid_input)

        lambda_label = QLabel("Lambda Coordinates (voxel): ")
        self.lambda_input1 = QLineEdit("0")
        self.lambda_input1.setValidator(valid_input)
        self.lambda_input2 = QLineEdit("0")
        self.lambda_input2.setValidator(valid_input)
        self.lambda_input3 = QLineEdit("0")
        self.lambda_input3.setValidator(valid_input)

        vox_size_label = QLabel("Voxel Size (um): ")
        self.vox_size_input1 = QLineEdit("0")
        self.vox_size_input1.setValidator(float_input)

        factor_label = QLabel("Factor (voxel): ")
        self.factor_input1 = QLineEdit("2")
        self.factor_input1.setValidator(QIntValidator(2, 99999))

        dim_selector_label = QLabel("CSys Selector: ")

        self.x_axis_combo = QComboBox()
        self.x_axis_combo.addItem("x-axis: ")
        self.x_axis_combo.addItems(self.selector_vals)
        self.x_axis_combo.currentIndexChanged.connect(
            lambda: self.dim_combo_changed("x")
        )

        self.y_axis_combo = QComboBox()
        self.y_axis_combo.addItem("y-axis: ")
        self.y_axis_combo.addItems(self.selector_vals)
        self.y_axis_combo.currentIndexChanged.connect(
            lambda: self.dim_combo_changed("y")
        )

        self.z_axis_combo = QComboBox()
        self.z_axis_combo.addItem("z-axis: ")
        self.z_axis_combo.addItems(self.selector_vals)
        self.z_axis_combo.currentIndexChanged.connect(
            lambda: self.dim_combo_changed("z")
        )

        self.process_btn = QPushButton("Start Process")
        self.process_btn.setAutoDefault(False)
        self.process_btn.setFocus(False)
        self.process_info = QLabel(
            "The whole process takes some time. \n"
            "This window will be closed automatically when processing finished."
        )

        self.progress = QProgressBar(self)
        self.progress.setMinimumWidth(100)
        self.progress.setTextVisible(False)
        self.progress_label = QLabel("0 %")
        self.progress_label.setFixedWidth(50)

        progress_wrap = QFrame()
        pw_layout = QHBoxLayout(progress_wrap)
        pw_layout.setSpacing(5)
        pw_layout.setContentsMargins(0, 0, 0, 0)
        pw_layout.setAlignment(Qt.AlignRight)
        pw_layout.addWidget(self.progress)
        pw_layout.addWidget(self.progress_label)

        layout.addWidget(data_label, 0, 0, 1, 1)
        layout.addWidget(self.data_btn, 0, 1, 1, 1)
        layout.addWidget(self.data_line, 0, 2, 1, 2)
        layout.addWidget(seg_label, 1, 0, 1, 1)
        layout.addWidget(self.seg_btn, 1, 1, 1, 1)
        layout.addWidget(self.seg_line, 1, 2, 1, 2)
        layout.addWidget(mask_label, 2, 0, 1, 1)
        layout.addWidget(self.mask_btn, 2, 1, 1, 1)
        layout.addWidget(self.mask_line, 2, 2, 1, 2)
        layout.addWidget(labinf_label, 3, 0, 1, 1)
        layout.addWidget(self.labinf_btn, 3, 1, 1, 1)
        layout.addWidget(self.labinf_line, 3, 2, 1, 2)
        layout.addWidget(bregma_label, 4, 0, 1, 1)
        layout.addWidget(self.bregma_input1, 4, 1, 1, 1)
        layout.addWidget(self.bregma_input2, 4, 2, 1, 1)
        layout.addWidget(self.bregma_input3, 4, 3, 1, 1)
        layout.addWidget(lambda_label, 5, 0, 1, 1)
        layout.addWidget(self.lambda_input1, 5, 1, 1, 1)
        layout.addWidget(self.lambda_input2, 5, 2, 1, 1)
        layout.addWidget(self.lambda_input3, 5, 3, 1, 1)
        layout.addWidget(vox_size_label, 6, 0, 1, 1)
        layout.addWidget(self.vox_size_input1, 6, 1, 1, 1)
        layout.addWidget(factor_label, 6, 2, 1, 1)
        layout.addWidget(self.factor_input1, 6, 3, 1, 1)

        layout.addWidget(dim_selector_label, 7, 0, 1, 1)
        layout.addWidget(self.x_axis_combo, 7, 1, 1, 1)
        layout.addWidget(self.y_axis_combo, 7, 2, 1, 1)
        layout.addWidget(self.z_axis_combo, 7, 3, 1, 1)

        layout.addWidget(self.process_info, 8, 0, 1, 4)
        layout.addWidget(self.process_btn, 9, 0, 1, 4)
        layout.addWidget(progress_wrap, 10, 0, 1, 4)

        # connect all buttons
        self.data_btn.clicked.connect(self.get_data_file)
        self.seg_btn.clicked.connect(self.get_seg_file)
        self.mask_btn.clicked.connect(self.get_mask_file)
        self.labinf_btn.clicked.connect(self.get_info_file)
        self.process_btn.clicked.connect(self.process_data_called)
        self.bregma_input1.textChanged.connect(self.bregma_input1_changed)
        self.bregma_input2.textChanged.connect(self.bregma_input2_changed)
        self.bregma_input3.textChanged.connect(self.bregma_input3_changed)
        self.lambda_input1.textChanged.connect(self.lambda_input1_changed)
        self.lambda_input2.textChanged.connect(self.lambda_input2_changed)
        self.lambda_input3.textChanged.connect(self.lambda_input3_changed)
        self.vox_size_input1.textChanged.connect(self.vox_size_input_changed)
        self.factor_input1.textChanged.connect(self.factor_input_changed)
        self.factor_input1.editingFinished.connect(self.factor_input_changed)

    def dim_combo_changed(self, ax):
        if not self.info_flag:
            self.process_info.setText("")
        if ax == "x":
            self.directions[0] = self.x_axis_combo.currentIndex()
        elif ax == "y":
            self.directions[1] = self.y_axis_combo.currentIndex()
        else:
            self.directions[2] = self.z_axis_combo.currentIndex()

        if np.all(np.ravel(self.directions) != 0):
            dir_groups = self.group_maps[np.array(self.directions) - 1]

            dir_goal = ["Post. --> Ant.", "Inf. --> Sup.", "L.H. --> R.H."]
            direction_change = [False, False, False]
            if self.x_axis_combo.currentText() not in dir_goal:
                direction_change[0] = True
            if self.y_axis_combo.currentText() not in dir_goal:
                direction_change[1] = True
            if self.z_axis_combo.currentText() not in dir_goal:
                direction_change[2] = True

            transpose_order = dir_groups - 1
            transpose_order = transpose_order.tolist()

            if transpose_order == [1, 2, 0]:
                inv_transpose_order = [2, 0, 1]
            elif transpose_order == [2, 0, 1]:
                inv_transpose_order = [1, 2, 0]
            else:
                inv_transpose_order = transpose_order.copy()

            self.axis_info = {
                "to_HERBS": tuple(inv_transpose_order),
                "from_HERBS": tuple(transpose_order),
                "direction_change": tuple(direction_change),
            }

            print(self.axis_info)

    def get_folder_path(self, file_path):
        self.folder_path = os.path.dirname(file_path)

    def get_data_file(self):
        file_options = QFileDialog.Options()
        file_options |= QFileDialog.DontUseNativeDialog
        dlg = QFileDialog()
        if self.folder_path is not None:
            data_path = dlg.getOpenFileName(
                self, "Select Atlas Volume File", self.folder_path, options=file_options
            )
        else:
            data_path = dlg.getOpenFileName(
                self, "Select Atlas Volume File", options=file_options
            )
            self.get_folder_path(data_path[0])
        self.data_local = os.path.basename(data_path[0])
        self.data_line.setText(self.data_local)

    def get_seg_file(self):
        file_options = QFileDialog.Options()
        file_options |= QFileDialog.DontUseNativeDialog
        dlg = QFileDialog()
        if self.folder_path is not None:
            seg_path = dlg.getOpenFileName(
                self, "Select Segmentation File", self.folder_path, options=file_options
            )
        else:
            seg_path = dlg.getOpenFileName(
                self, "Select Segmentation File", options=file_options
            )
            self.get_folder_path(seg_path[0])
        self.segmentation_local = os.path.basename(seg_path[0])
        self.seg_line.setText(self.segmentation_local)

    def get_mask_file(self):
        file_options = QFileDialog.Options()
        file_options |= QFileDialog.DontUseNativeDialog
        dlg = QFileDialog()
        if self.folder_path is not None:
            mask_path = dlg.getOpenFileName(
                self, "Select Mask File", self.folder_path, options=file_options
            )
        else:
            mask_path = dlg.getOpenFileName(
                self, "Select Mask File", options=file_options
            )
            self.get_folder_path(mask_path[0])
        self.mask_local = os.path.basename(mask_path[0])
        self.mask_line.setText(self.mask_local)

    def get_info_file(self):
        file_options = QFileDialog.Options()
        file_options |= QFileDialog.DontUseNativeDialog
        dlg = QFileDialog()
        if self.folder_path is not None:
            info_path = dlg.getOpenFileName(
                self, "Select Label File", self.folder_path, options=file_options
            )
        else:
            info_path = dlg.getOpenFileName(
                self, "Select Label File", options=file_options
            )
            self.get_folder_path(info_path[0])
        self.label_local = os.path.basename(info_path[0])
        self.labinf_line.setText(os.path.basename(self.label_local))

    def bregma_input1_changed(self, text):
        if text == "":
            return
        self.bregma_coord[0] = int(text)

    def bregma_input2_changed(self, text):
        if text == "":
            return
        self.bregma_coord[1] = int(text)

    def bregma_input3_changed(self, text):
        if text == "":
            return
        self.bregma_coord[2] = int(text)

    def lambda_input1_changed(self, text):
        if text == "":
            return
        self.lambda_coord[0] = int(text)

    def lambda_input2_changed(self, text):
        if text == "":
            return
        self.lambda_coord[1] = int(text)

    def lambda_input3_changed(self, text):
        if text == "":
            return
        self.lambda_coord[2] = int(text)

    def vox_size_input_changed(self, text):
        if text == "":
            return
        self.voxel_size = float(text)

    def factor_input_changed(self, text):
        if text == "":
            return
        self.factor_val = int(text)

    def check_empty_file(self):
        if self.data_local is None:
            msg = "Please select atlas data."
            return msg
        if self.segmentation_local is None:
            msg = "Please select segmentation file."
            return msg
        if self.label_local is None:
            msg = "Please select label information file."
            return msg
        return None

    def process_data_called(self):
        if np.any(np.array(self.directions) == 0):
            self.info_flag = False
            self.process_info.setText("Please select directions for each axis.")
            return
        dir_groups = self.group_maps[np.array(self.directions) - 1]
        if len(np.unique(dir_groups)) != 3:
            self.info_flag = False
            self.process_info.setText("Axis directions can not duplicat.")
            return

        msg = self.check_empty_file()
        if msg is not None:
            self.process_info.setText("Please select file path.")
            return

        dir_goal = ["Post. --> Ant.", "Inf. --> Sup.", "L.H. --> R.H."]
        direction_change = [False, False, False]
        if self.x_axis_combo.currentText() not in dir_goal:
            direction_change[0] = True
        if self.y_axis_combo.currentText() not in dir_goal:
            direction_change[1] = True
        if self.z_axis_combo.currentText() not in dir_goal:
            direction_change[2] = True

        transpose_order = dir_groups - 1

        self.process_btn.setVisible(False)
        self.process_info.setText(
            "The whole process takes some time. \n"
            "This window will be closed automatically when processing finished."
        )

        self.thread = QThread()
        self.worker = CustomerAtlasWorker()
        self.worker.set_data(
            self.folder_path,
            self.data_local,
            self.segmentation_local,
            self.label_local,
            self.axis_info,
            self.bregma_coord,
            self.voxel_size,
            self.mask_local,
            self.factor_val,
        )
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.on_finish)
        self.worker.error_occur.connect(self.report_error)
        # self.worker.finished.connect(self.worker.deleteLater)
        self.thread.finished.connect(self.thread.deleteLater)
        self.worker.progress.connect(self.report_progress)
        self.thread.start()

    def report_error(self, msg):
        self.thread.quit()
        self.process_btn.setVisible(True)
        self.progress.setValue(int(0))
        self.progress_label.setText("%.02f %%" % 0)
        self.process_info.setText(msg)

    def report_progress(self, val):
        val = np.round(val, 2)
        self.progress.setValue(int(val))
        # self.progress.setFormat("%.02f %%" % val)
        self.progress_label.setText("%.02f %%" % val)

    def on_finish(self):
        self.thread.quit()
        self.close()



from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np
from pyqtgraph.Qt import QtGui, QtCore


class ToolBox(QObject):

    def __init__(self):

        QObject.__init__(self)

        self.pencil_mask = np.ones((3, 3), 'uint8')
        self.is_closed = False
        self.multi_shanks = False

        self.remove_inside = True

        self.merge_sites = False

        # action version
        self.add_atlas = QAction(QIcon('icons/toolbar/atlas_icon.png'), 'Upload Previous Loaded Volume Atlas', self)
        self.add_image_stack = QAction(QIcon('icons/toolbar/image_icon.svg'), 'Upload Histological Image', self)

        self.vis2 = QAction(QIcon('icons/toolbar/two_window.png'), 'Show 2 Windows', self)
        self.vis4 = QAction(QIcon('icons/toolbar/window4.png'), 'Show 4 Windows', self)
        self.vis3 = QAction(QIcon('icons/toolbar/window3.png'), 'Show 3 Windows', self)

        self.toa_btn_off_icon = QIcon('icons/toolbar/toa.svg')
        self.toa_btn_on_icon = QIcon('icons/toolbar/toa_delete.svg')

        self.toh_btn_off_icon = QIcon('icons/toolbar/toh.svg')
        self.toh_btn_on_icon = QIcon('icons/toolbar/toh_delete.svg')

        self.toa_btn = QAction(QIcon('icons/toolbar/toa.svg'), 'Transform to Atlas Slice Window', self)
        self.toh_btn = QAction(QIcon('icons/toolbar/toh.svg'), 'Transform to Histologist Image Window', self)
        self.check_btn = QAction(QIcon('icons/toolbar/accept.svg'), 'Accept and Transfer', self)

        lasso_btn = QAction(QIcon('icons/toolbar/lasso.svg'), 'Polygon Lasso', self)
        lasso_btn.setCheckable(True)

        magic_wand_btn = QAction(QIcon('icons/toolbar/magic-wand.svg'), 'Magic Wand', self)
        magic_wand_btn.setCheckable(True)

        pencil_btn = QAction(QIcon('icons/toolbar/pencil.svg'), 'Pencil', self)
        pencil_btn.setCheckable(True)

        eraser_btn = QAction(QIcon('icons/toolbar/eraser.svg'), 'Eraser', self)
        eraser_btn.setCheckable(True)

        mask_btn = QAction(QIcon('icons/toolbar/mask.svg'), 'Mask Maker', self)
        mask_btn.setCheckable(True)

        probe_btn = QAction(QIcon('icons/toolbar/probe.svg'), 'Probe Marker', self)
        probe_btn.setCheckable(True)

        triang_btn = QAction(QIcon('icons/toolbar/triangulation.svg'), 'Triangulation', self)
        triang_btn.setCheckable(True)

        loc_btn = QAction(QIcon('icons/toolbar/location.svg'), 'Cell Selector', self)
        loc_btn.setCheckable(True)

        ruler_btn = QAction(QIcon('icons/toolbar/ruler.svg'), 'Ruler', self)
        ruler_btn.setCheckable(True)

        self.checkable_btn_dict = {'ruler_btn': ruler_btn,
                                   'pencil_btn': pencil_btn,
                                   'eraser_btn': eraser_btn,
                                   'lasso_btn': lasso_btn,
                                   'magic_wand_btn': magic_wand_btn,
                                   'probe_btn': probe_btn,
                                   'triang_btn': triang_btn,
                                   'loc_btn': loc_btn}

        self.toolbox_btn_keys = list(self.checkable_btn_dict.keys())

        # for ruler_btn, ruler wrap
        ruler_color_label = QLabel('Color:')
        self.ruler_color_btn = pg.ColorButton(padding=0)
        self.ruler_color_btn.setColor('yellow')
        self.ruler_color_btn.setFixedSize(60, 15)
        ruler_width_label = QLabel('Size:')
        self.ruler_width_slider = QSlider(Qt.Horizontal)
        self.ruler_width_slider.setFixedWidth(100)
        self.ruler_width_slider.setMinimum(1)
        self.ruler_width_slider.setMaximum(5)
        self.ruler_width_slider.setValue(3)
        self.ruler_width_slider.sliderMoved.connect(self.change_ruler_slider)

        self.ruler_size_valt = QLineEdit()
        self.ruler_size_valt.setFixedSize(50, 24)
        self.ruler_size_valt.setAlignment(Qt.AlignLeft)
        self.ruler_size_valt.setValidator(QIntValidator(1, 5))
        self.ruler_size_valt.setText('3')

        self.ruler_length_label = QLabel('Length: ')

        self.ruler_wrap = QFrame()
        ruler_layout = QHBoxLayout(self.ruler_wrap)
        ruler_layout.setAlignment(Qt.AlignVCenter)
        ruler_layout.setContentsMargins(0, 0, 0, 0)
        ruler_layout.setSpacing(10)
        ruler_layout.addWidget(ruler_color_label)
        ruler_layout.addWidget(self.ruler_color_btn)
        ruler_layout.addSpacing(15)
        ruler_layout.addWidget(ruler_width_label)
        ruler_layout.addWidget(self.ruler_width_slider)
        ruler_layout.addWidget(self.ruler_size_valt)
        ruler_layout.addSpacing(25)
        ruler_layout.addWidget(self.ruler_length_label)
        ruler_layout.addStretch(1)

        # for magic wand, magic wand wrap
        magic_color_label = QLabel('Color:')
        self.magic_color_btn = pg.ColorButton(padding=0)
        self.magic_color_btn.setFixedSize(60, 15)
        self.magic_color_btn.setColor(QColor(255, 0, 255, 255))
        magic_tol_label = QLabel('Tolerance:')
        self.magic_tol_val = QLineEdit()
        self.magic_tol_val.setFixedWidth(40)
        # self.magic_tol_val.setAlignment(Qt.AlignLeft)
        self.magic_tol_val.setValidator(QIntValidator())
        self.magic_tol_val.setText('0')

        self.magic_wand_kernel = QComboBox()
        self.magic_wand_kernel.setFixedSize(150, 20)
        self.magic_wand_kernel.addItems(["Kernel", "Rectangular", "Elliptical", "Cross-shaped"])

        magic_wand_ksize_label = QLabel('Size:')
        self.magic_wand_ksize = QSpinBox()
        self.magic_wand_ksize.setFixedSize(80, 20)
        self.magic_wand_ksize.setValue(0)
        self.magic_wand_ksize.setMinimum(0)

        self.magic_wand_virus_register = QPushButton()
        self.magic_wand_virus_register.setFocusPolicy(Qt.NoFocus)
        self.magic_wand_virus_register.setIcon(QIcon('icons/toolbar/virus_register.svg'))
        self.magic_wand_virus_register.setIconSize(QSize(20, 20))
        self.magic_wand_bnd_register = QPushButton()
        self.magic_wand_bnd_register.setFocusPolicy(Qt.NoFocus)
        self.magic_wand_bnd_register.setIcon(QIcon('icons/toolbar/boundary_register.svg'))
        self.magic_wand_bnd_register.setIconSize(QSize(20, 20))

        self.magic_wand_wrap = QFrame()
        # self.magic_wand_wrap.setFixedHeight(32)
        magic_wand_layout = QHBoxLayout(self.magic_wand_wrap)
        magic_wand_layout.setAlignment(Qt.AlignVCenter)
        magic_wand_layout.setContentsMargins(0, 0, 0, 0)
        magic_wand_layout.setSpacing(10)
        magic_wand_layout.addWidget(magic_color_label)
        magic_wand_layout.addWidget(self.magic_color_btn)
        magic_wand_layout.addSpacing(10)
        magic_wand_layout.addWidget(magic_tol_label)
        magic_wand_layout.addWidget(self.magic_tol_val)
        magic_wand_layout.addWidget(self.magic_wand_kernel)
        magic_wand_layout.addWidget(magic_wand_ksize_label)
        magic_wand_layout.addWidget(self.magic_wand_ksize)
        magic_wand_layout.addWidget(self.magic_wand_virus_register)
        magic_wand_layout.addWidget(self.magic_wand_bnd_register)
        magic_wand_layout.addStretch(1)
        # magic_wand_layout.insertStretch(1, -1)

        # for pencil_btn, pencil wrap
        pencil_size_label = QLabel('Size:')
        self.pencil_size_slider = QSlider(Qt.Horizontal)
        self.pencil_size_slider.setFixedWidth(100)
        self.pencil_size_slider.setMinimum(1)
        self.pencil_size_slider.setMaximum(5)
        self.pencil_size_slider.setValue(2)
        self.pencil_size_slider.sliderMoved.connect(self.change_pencil_slider)

        self.pencil_size_valt = QLineEdit()
        self.pencil_size_valt.setFixedSize(50, 24)
        self.pencil_size_valt.setAlignment(Qt.AlignLeft)
        self.pencil_size_valt.setValidator(QIntValidator(1, 5))
        self.pencil_size_valt.setText('2')

        pencil_color_label = QLabel('Color:')
        self.pencil_color_btn = pg.ColorButton(padding=0)
        self.pencil_color_btn.setFixedSize(60, 15)
        self.pencil_color_btn.setColor(QColor(255, 102, 0))

        pencil_path_label = QLabel('Path Type:')
        self.pencil_path_btn = QPushButton()
        self.pencil_path_btn.setCheckable(True)
        da_icon = QIcon()
        da_icon.addPixmap(QPixmap("icons/toolbar/closed_path.svg"), QIcon.Normal, QIcon.On)
        da_icon.addPixmap(QPixmap("icons/toolbar/open_path.svg"), QIcon.Normal, QIcon.Off)
        self.pencil_path_btn.setIcon(da_icon)
        self.pencil_path_btn.setIconSize(QSize(20, 20))
        self.pencil_path_btn.clicked.connect(self.pencil_path_btn_clicked)

        self.pencil_wrap = QFrame()
        pencil_layout = QHBoxLayout(self.pencil_wrap)
        pencil_layout.setAlignment(Qt.AlignVCenter)
        pencil_layout.setContentsMargins(0, 0, 0, 0)
        pencil_layout.setSpacing(10)
        pencil_layout.addWidget(pencil_color_label)
        pencil_layout.addWidget(self.pencil_color_btn)
        pencil_layout.addSpacing(15)
        pencil_layout.addWidget(pencil_size_label)
        pencil_layout.addWidget(self.pencil_size_slider)
        pencil_layout.addWidget(self.pencil_size_valt)
        pencil_layout.addSpacing(15)
        pencil_layout.addWidget(pencil_path_label)
        pencil_layout.addWidget(self.pencil_path_btn)
        pencil_layout.addStretch(1)

        # for eraser_btn, eraser wrap
        eraser_color_label = QLabel('Color:')
        self.eraser_color_btn = pg.ColorButton(padding=0)
        self.eraser_color_btn.setColor('red')
        self.eraser_color_btn.setFixedSize(60, 15)
        eraser_size_label = QLabel('Size:')
        self.eraser_size_slider = QSlider(Qt.Horizontal)
        self.eraser_size_slider.setFixedWidth(100)
        self.eraser_size_slider.setMinimum(1)
        self.eraser_size_slider.setMaximum(500)
        self.eraser_size_slider.setValue(20)
        self.eraser_size_slider.valueChanged.connect(self.change_eraser_slider)

        self.eraser_size_valt = QLineEdit()
        self.eraser_size_valt.setFixedSize(50, 24)
        self.eraser_size_valt.setAlignment(Qt.AlignLeft)
        self.eraser_size_valt.setValidator(QIntValidator(1, 500))
        self.eraser_size_valt.setText('20')
        self.eraser_size_valt.textChanged.connect(self.change_eraser_val)

        self.eraser_wrap = QFrame()
        eraser_layout = QHBoxLayout(self.eraser_wrap)
        eraser_layout.setAlignment(Qt.AlignVCenter)
        eraser_layout.setContentsMargins(0, 0, 0, 0)
        eraser_layout.setSpacing(10)
        eraser_layout.addWidget(eraser_color_label)
        eraser_layout.addWidget(self.eraser_color_btn)
        eraser_layout.addSpacing(15)
        eraser_layout.addWidget(eraser_size_label)
        eraser_layout.addWidget(self.eraser_size_slider)
        eraser_layout.addWidget(self.eraser_size_valt)
        eraser_layout.addStretch(1)

        # for lasso_btn, lasso wrap
        lasso_color_label = QLabel('Color:')
        self.lasso_color_btn = pg.ColorButton(padding=0)
        self.lasso_color_btn.setFixedSize(60, 15)
        self.lasso_color_btn.setColor(QColor(0, 255, 255))

        self.lasso_type_btn = QPushButton()
        self.lasso_type_btn.setCheckable(True)
        lt_icon = QIcon()
        lt_icon.addPixmap(QPixmap('icons/toolbar/outpart.svg'), QIcon.Normal, QIcon.On)
        lt_icon.addPixmap(QPixmap('icons/toolbar/inpart.svg'), QIcon.Normal, QIcon.Off)
        self.lasso_type_btn.setIcon(lt_icon)
        self.lasso_type_btn.setIconSize(QSize(20, 20))
        self.lasso_type_btn.clicked.connect(self.lasso_type_changed)

        self.lasso_wrap = QFrame()
        lasso_layout = QHBoxLayout(self.lasso_wrap)
        lasso_layout.setAlignment(Qt.AlignVCenter)
        lasso_layout.setContentsMargins(0, 0, 0, 0)
        lasso_layout.setSpacing(10)
        lasso_layout.addWidget(lasso_color_label)
        lasso_layout.addWidget(self.lasso_color_btn)
        lasso_layout.addSpacing(10)
        lasso_layout.addWidget(self.lasso_type_btn)
        lasso_layout.addStretch(1)

        # mask wrap
        kernel_size_label = QLabel('Kernel Size: ')
        self.kernel_size_slider = QSlider(Qt.Horizontal)
        self.kernel_size_slider.setFixedWidth(100)
        self.kernel_size_slider.setMinimum(1)
        self.kernel_size_slider.setMaximum(50)
        self.kernel_size_slider.setValue(10)

        self.mask_wrap = QFrame()
        mask_layout = QHBoxLayout(self.mask_wrap)
        mask_layout.setAlignment(Qt.AlignVCenter)
        mask_layout.setContentsMargins(0, 0, 0, 0)
        mask_layout.setSpacing(10)
        mask_layout.addWidget(kernel_size_label)
        mask_layout.addWidget(self.kernel_size_slider)
        mask_layout.addStretch(1)

        # for probe_btn, probe wrap
        self.multi_prb_btn = QPushButton()
        self.multi_prb_btn.setCheckable(True)
        mp_icon = QIcon()
        mp_icon.addPixmap(QPixmap("icons/toolbar/multi_pencil.svg"), QIcon.Normal, QIcon.On)
        mp_icon.addPixmap(QPixmap("icons/toolbar/single_pencil.svg"), QIcon.Normal, QIcon.Off)
        self.multi_prb_btn.setIcon(mp_icon)
        self.multi_prb_btn.setIconSize(QSize(20, 20))
        self.multi_prb_btn.setToolTip('Multi-shanks Switch')

        probe_color_label = QLabel('Color:')
        probe_color_label.setToolTip('Probe Color')
        self.probe_color_btn = pg.ColorButton(padding=0)
        self.probe_color_btn.setColor(QColor(0, 0, 255))
        self.probe_color_btn.setFixedSize(60, 15)
        probe_type_label = QLabel('Type:')
        probe_type_label.setToolTip('Probe type')
        self.probe_type_combo = QComboBox()
        self.probe_type_combo.setFixedSize(150, 20)
        self.probe_type_combo.addItems(["Neuropixel 1.0", "Neuropixel 2.0", "Linear-silicon", "Tetrode"])
        self.probe_type_combo.setCurrentText("Neuropixel 1.0")

        self.merge_sites_btn = QPushButton()
        self.merge_sites_btn.setCheckable(True)
        ms_icon = QIcon()
        ms_icon.addPixmap(QPixmap("icons/toolbar/line_sites.svg"), QIcon.Normal, QIcon.On)
        ms_icon.addPixmap(QPixmap("icons/toolbar/separate_sites.svg"), QIcon.Normal, QIcon.Off)
        self.merge_sites_btn.setIcon(ms_icon)
        self.merge_sites_btn.setIconSize(QSize(20, 20))
        self.merge_sites_btn.setToolTip('Merge sites')
        self.merge_sites_btn.clicked.connect(self.merge_sites_btn_clicked)

        site_face_label = QLabel('Site-face:')
        site_face_label.setToolTip('Sites Face Direction')
        self.pre_site_face_combo = QComboBox()
        self.pre_site_face_combo.setFixedSize(100, 20)
        self.pre_site_face_combo.addItems(["Out", "In", "Left", "Right"])

        self.after_site_face_combo = QComboBox()
        self.after_site_face_combo.setFixedSize(100, 20)
        self.after_site_face_combo.addItems(["Up", "Down", "Left", "Right"])
        self.after_site_face_combo.setVisible(False)

        self.linear_silicon_list = QPushButton()
        self.linear_silicon_list.setFocusPolicy(Qt.NoFocus)
        self.linear_silicon_list.setIcon(QIcon('icons/toolbar/list.svg'))
        self.linear_silicon_list.setIconSize(QSize(20, 20))
        self.linear_silicon_list.setVisible(False)
        self.linear_silicon_list.setToolTip('Linear Silicon Designer')

        self.probe_wrap = QFrame()
        probe_layout = QHBoxLayout(self.probe_wrap)
        probe_layout.setAlignment(Qt.AlignVCenter)
        probe_layout.setContentsMargins(0, 0, 0, 0)
        probe_layout.setSpacing(10)
        probe_layout.addWidget(self.multi_prb_btn)
        probe_layout.addSpacing(20)
        probe_layout.addWidget(probe_color_label)
        probe_layout.addWidget(self.probe_color_btn)
        probe_layout.addSpacing(10)
        probe_layout.addWidget(probe_type_label)
        probe_layout.addWidget(self.probe_type_combo)
        probe_layout.addSpacing(10)
        probe_layout.addWidget(self.merge_sites_btn)
        probe_layout.addWidget(site_face_label)
        probe_layout.addWidget(self.pre_site_face_combo)
        probe_layout.addWidget(self.after_site_face_combo)
        probe_layout.addStretch(1)
        probe_layout.addWidget(self.linear_silicon_list)
        probe_layout.addSpacing(10)

        # for triang_btn, triang wrap
        triang_color_label = QLabel('Color:')
        self.triang_color_btn = pg.ColorButton(padding=0)
        self.triang_color_btn.setFixedSize(60, 15)
        bound_pnts_num_label = QLabel('Points Number:')
        self.bound_pnts_num = QLineEdit()
        self.bound_pnts_num.setFixedSize(50, 24)
        self.bound_pnts_num.setAlignment(Qt.AlignLeft)
        self.bound_pnts_num.setMaxLength(2)
        self.bound_pnts_num.setValidator(QIntValidator())
        self.bound_pnts_num.setText('2')
        self.triang_vis_btn = QPushButton()
        self.triang_vis_btn.setCheckable(True)
        vis_icon = QIcon()
        vis_icon.addPixmap(QPixmap("icons/toolbar/eye.svg"), QIcon.Normal, QIcon.On)
        vis_icon.addPixmap(QPixmap("icons/toolbar/eye_closed.svg"), QIcon.Normal, QIcon.Off)
        self.triang_vis_btn.setIcon(vis_icon)
        self.triang_vis_btn.setIconSize(QSize(20, 20))
        self.triang_match_bnd = QPushButton()
        self.triang_match_bnd.setIcon(QIcon('icons/toolbar/matchbnd.svg'))
        self.triang_match_bnd.setIconSize(QSize(20, 20))

        self.triang_wrap = QFrame()
        self.triang_wrap.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        triang_layout = QHBoxLayout(self.triang_wrap)
        triang_layout.setAlignment(Qt.AlignVCenter)
        triang_layout.setContentsMargins(0, 0, 0, 0)
        triang_layout.setSpacing(10)
        triang_layout.addWidget(self.triang_vis_btn)
        triang_layout.addSpacing(10)
        triang_layout.addWidget(self.triang_match_bnd)
        triang_layout.addSpacing(10)
        triang_layout.addWidget(triang_color_label)
        triang_layout.addWidget(self.triang_color_btn)
        triang_layout.addSpacing(10)
        triang_layout.addWidget(bound_pnts_num_label)
        triang_layout.addWidget(self.bound_pnts_num)

        triang_layout.addStretch(1)

        # cell count wrap
        cell_color_label = QLabel('Color:')
        self.cell_color_btn = pg.ColorButton(padding=0)
        self.cell_color_btn.setColor(QColor(0, 255, 0))
        self.cell_color_btn.setFixedSize(60, 15)

        self.cell_count_label_list = [QLabel('Count: ')]
        self.cell_count_val_list = [QLabel('0')]

        for i in range(4):
            self.cell_count_label_list.append(QLabel('Count {}: '.format(i+1)))
            self.cell_count_val_list.append(QLabel('0'))

        for i in range(5):
            self.cell_count_label_list[i].setVisible(False)
            self.cell_count_val_list[i].setVisible(False)

        self.cell_selector_btn = QPushButton()
        self.cell_selector_btn.setToolTip('select cells manually')
        self.cell_selector_btn.setCheckable(True)
        selector_icon = QIcon()
        selector_icon.addPixmap(QPixmap("icons/toolbar/cell_select_not.svg"), QIcon.Normal, QIcon.Off)
        selector_icon.addPixmap(QPixmap("icons/toolbar/cell_select.svg"), QIcon.Normal, QIcon.On)
        self.cell_selector_btn.setIcon(selector_icon)
        self.cell_selector_btn.setIconSize(QSize(20, 20))

        self.cell_aim_btn = QPushButton()
        self.cell_aim_btn.setToolTip('select single cell body area')
        self.cell_aim_btn.setCheckable(True)
        aim_icon = QIcon()
        aim_icon.addPixmap(QPixmap("icons/toolbar/aim_not.svg"), QIcon.Normal, QIcon.Off)
        aim_icon.addPixmap(QPixmap("icons/toolbar/aim.svg"), QIcon.Normal, QIcon.On)
        self.cell_aim_btn.setIcon(aim_icon)
        self.cell_aim_btn.setIconSize(QSize(20, 20))
        self.cell_radar_btn = QPushButton()
        self.cell_radar_btn.setToolTip('automatically search for similar cell bodies')
        self.cell_radar_btn.setIcon(QIcon('icons/toolbar/radar.svg'))
        self.cell_radar_btn.setIconSize(QSize(20, 20))

        self.cell_count_wrap = QFrame()
        cell_layout = QHBoxLayout(self.cell_count_wrap)
        cell_layout.setAlignment(Qt.AlignVCenter)
        cell_layout.setContentsMargins(0, 0, 0, 0)
        cell_layout.setSpacing(10)
        cell_layout.addWidget(cell_color_label)
        cell_layout.addWidget(self.cell_color_btn)
        cell_layout.addSpacing(15)
        cell_layout.addWidget(self.cell_selector_btn)
        cell_layout.addSpacing(10)
        cell_layout.addWidget(self.cell_aim_btn)
        cell_layout.addSpacing(10)
        cell_layout.addWidget(self.cell_radar_btn)
        cell_layout.addSpacing(10)
        cell_layout.addWidget(self.cell_count_label_list[0])
        cell_layout.addWidget(self.cell_count_val_list[0])
        for i in range(1, 5):
            cell_layout.addWidget(self.cell_count_label_list[i])
            cell_layout.addWidget(self.cell_count_val_list[i])
            cell_layout.addSpacing(10)
        cell_layout.addStretch(1)

        # separator
        self.sep_label = QLabel()
        self.sep_label.setPixmap(QPixmap('icons/toolbar/handle.png'))

        # circle
        r = int(self.eraser_size_valt.text())
        angle = np.deg2rad(np.arange(0, 360, 1))
        x = np.cos(angle)
        y = np.sin(angle)
        self.original_circle = np.vstack([x, y]).T
        self.circle = self.original_circle * r

        # triangle style
        self.tri_line_style = pg.mkPen(color=(128, 128, 128), width=0.5, style=Qt.DashLine)

        # ---------------------------- define all cursor shape
        # self.eraser_cursor = QCursor(QPixmap('icons/eraser_cursor.png'), hotX=7, hotY=27)

    def update_cell_count_label(self, cell_count_list):
        for layer_index in range(5):
            self.update_single_cell_count_label(cell_count_list, layer_index)

    def update_single_cell_count_label(self, cell_count_list, layer_index):
        self.cell_count_val_list[layer_index].setText(str(cell_count_list[layer_index]))

    def lasso_type_changed(self):
        if self.lasso_type_btn.isChecked():
            self.remove_inside = False
        else:
            self.remove_inside = True

    def pencil_path_btn_clicked(self):
        if self.pencil_path_btn.isChecked():
            self.is_closed = True
        else:
            self.is_closed = False

    def merge_sites_btn_clicked(self):
        if self.merge_sites_btn.isChecked():
            self.merge_sites = True
        else:
            self.merge_sites = False

    def change_pencil_slider(self):
        val = self.pencil_size_slider.value()
        self.pencil_size_valt.setText(str(val))

    def change_eraser_slider(self):
        val = self.eraser_size_slider.value()
        self.eraser_size_valt.setText(str(val))
        self.circle = self.original_circle * val

    def change_eraser_val(self):
        val = int(self.eraser_size_valt.text())
        self.eraser_size_slider.setValue(val)

    def change_ruler_slider(self):
        val = self.ruler_width_slider.value()
        self.ruler_size_valt.setText(str(val))

    def get_tool_data(self):
        data = {'pencil_color': self.pencil_color_btn.color(),
                'pencil_size': self.pencil_size_valt.text(),
                'magic_wand_color': self.magic_color_btn.color(),
                'magic_wand_tol': self.magic_tol_val.text(),
                'magic_wand_kernel': self.magic_wand_kernel.currentText(),
                'magic_wand_ksize': self.magic_wand_ksize.value(),
                'probe_color': self.probe_color_btn.color(),
                'cell_color': self.cell_color_btn.color(),
                'is_closed': self.is_closed}
        return data

    def set_tool_data(self, data):
        self.pencil_color_btn.setColor(data['pencil_color'])
        self.pencil_size_valt.setText(str(data['pencil_size']))
        self.pencil_path_btn.setChecked(data['is_closed'])
        self.magic_color_btn.setColor(data['magic_wand_color'])
        self.magic_tol_val.setText(str(data['magic_wand_tol']))
        self.magic_wand_ksize.setValue(data['magic_wand_ksize'])
        self.magic_wand_kernel.setCurrentText(data['magic_wand_kernel'])
        self.probe_color_btn.setColor(data['probe_color'])
        self.cell_color_btn.setColor(data['cell_color'])


    def multi_prb_status_changed(self):
        if self.multi_prb_btn.isChecked():
            self.multi_shanks = True
        else:
            self.multi_shanks = False

















from PyQt5.QtWidgets import *


class AboutHERBSWindow(QMessageBox):
    def __init__(self):
        super().__init__()

        self.setIcon(QMessageBox.NoIcon)
        self.setWindowTitle("About HERBS")
        self.setText("HERBS is aiming to provide a pleasant platform "
                     "for histological image registration in neuroscience. \n"
                     "\n"
                     "If you have any requests or questions ---- \n"
                     "\n"
                     "Please contact maintainers: \n"
                     "jingyi.g.fuglstad@gmail.com \n"
                     "\n"
                     "Or leave an issue/discussion on HERBS GitHub: \n"
                     "https://github.com/JingyiGF/HERBS \n"
                     "\n"
                     "Please always read the Update Log after updating: \n"
                     "https://github.com/JingyiGF/HERBS/blob/main/UpdateLog.md")
        self.setStandardButtons(QMessageBox.Close)
